### 
### OBEX protocol folder-listing xml-parser
### 
### <?xml version="1.0"?>
### <!DOCTYPE folder-listing SYSTEM "obex-folder-listing.dtd">
### <folder-listing version="1.0">
###    <file name="demotimeSheet.jad" size="301" 
###            modified="20030214T235606" user-perm="RWD" group-perm="R" />
###    <file name="demotimeSheet.jar" size="23949" 
###            modified="20030214T235614" user-perm="RWD" group-perm="R" />
###    <folder name="storage" modified="20030214T235644" 
###            user-perm="RWD" group-perm="W" />
### </folder-listing>
###
### History:
###   20031107: config params obexexe and obexdev removed - a wrapper
###             shell $OBEXDIR/obexftp.sh will be used in future - ger
###   20040504: parser reworked - newlines become ignored - ger
###   20040712: patch for Nokia 6230, provided by Michael Wikberg -
###             thank you Michael! - ger
###

package provide obexfile 0.4

namespace eval ObexFile {

  variable obex0104_patch 0

  variable array TAG
  set TAG(end) "/>"

  ### siemens M50   <?xml version="1.0"?>
  ### ericsson T610 <?xml version="1.0" encoding="UTF-8"?>
  set TAG(XML_identifier) \
      "<?xml version=\"1.0\"" 

  ### siemens M50 <!DOCTYPE folder-listing SYSTEM "obex-folder-listing.dtd">
  ### nokia 6230  <!DOCTYPE folder-listing SYSTEM "obex-folder-listing.dtd"
  ###              [ <!ATTLIST folder mem-type CDATA #IMPLIED> ]>
  set TAG(DTD_identifier) \
      "<!DOCTYPE folder-listing SYSTEM \"obex-folder-listing.dtd\""

  set TAG(folder-listbeg) \
      "<folder-listing version=\"1.0\">"
  set TAG(folder-listend) \
      "</folder-listing>"

  ### nokia 6230 specific <parent-folder />
  ###  ... does not appear in mmc-subdirectory ?!?!
  set TAG(parent-folder) \
      "<parent-folder "

  set TAG(file_begin) "<file"
  set TAG(file_end) $TAG(end)
  set TAG(folder_begin) "<folder"
  set TAG(folder_end) $TAG(end)

  variable separator [list " \t\n"]
  variable tmp_prefix "/tmp/otmp"

proc format_date { idate } {
  global ObexConfig
  if {![string compare $idate ""]} {return $idate}
  set date_format $ObexConfig(file,datefmt)
  set date_res ""
  set sl [string length $date_format]
  for {set i 0} {$i<$sl} {incr i} {
    set fmt_c [string index $date_format $i]
    set ufmt [string toupper $fmt_c]
    switch -- $ufmt {
      Y {  
        set pattern [string range $idate 0 3] ; ### 4 digit year
      }
      Z { 
        set pattern [string range $idate 2 3] ; ### 2 digit year
      } 
      M { 
        set pattern [string range $idate 4 5] ; ### 2 digit month
      } 
      D { 
        set pattern [string range $idate 6 7] ; ### 2 digit day
      } 
      H { 
        set pattern [string range $idate 9 10] ; ### 2 digit hour
      } 
      N { 
        set pattern [string range $idate 11 12] ; ### 2 digit minute
      } 
      S { 
        set pattern [string range $idate 13 14] ; ### 2 digit seconds
      } 
      default { set pattern $fmt_c }
    }
    if {[lsearch -exact [list z m d h n s] $fmt_c] != -1} {
      append date_res [expr 1 * $pattern] ; ### format to integer
    } else {
      append date_res $pattern
    }
  }
  return $date_res
}
proc extension { path } {
  set ext [file extension $path]
  if [string length $ext] {
    return [string range $ext 1 end]
  } else {
    return $ext
  }
}
proc get_filetype { stoken name } {
  variable TAG
  if {![string compare $stoken $TAG(folder_begin)]} {
    return folder
  } 
  return [string tolower [extension $name]]
}
proc name_sep { ch } {
  if {[string is alpha $ch]} {return 1}
  if {![string compare $ch "-"]} {return 1}
  return 0
}
proc att_value_list { arr fline } {
# debug_out "att_value_list $arr $fline"
  upvar $arr key
  set sl [string length $fline]
###
### extract named token (vname)
###
  while {$sl > 0} {
    set i 0
    set vname ""
    set ch1 [string index $fline $i]
    while {[name_sep $ch1]} {
      append vname $ch1
      set ch1 [string index $fline [incr i]]
# debug_out "ch1='$ch1' vname=$vname"
      if { $i >= $sl } break;
    }
    set fline [string range $fline [incr i] end]
    set fline [string trim $fline]
# debug_out "fline=$fline"
###
### only keyword, no value follows (token=value)?
###
# debug_out "ch1='$ch1'"
    set value ""
    if { $ch1 != "=" } {
      set value ""
      set key($vname) $value 
      set fline [string trimleft $fline]
      set sl [string length $fline]
      continue
    }
###
### value without quotes?
###
    set fline [string trim $fline]
    set ch1 [string index $fline [set i 0]]
# debug_out "$ch1 != '\"'"
    if { $ch1 != "\"" } {
      set value [lindex [split $fline] 0]
      set $fline [string range $fline [string length $value] end]
      set key($vname) $value 
      set fline [string trimleft $fline]
      set sl [string length $fline]
      continue
    }
###
### value with quotes (may include white spaces...)
###
    set sl [string length $fline]
    set ch1 [string index $fline [set i 1]]
# debug_out "ch1 == '$ch1'"
    while { $ch1 != "\"" } {
      append value $ch1
      set ch1 [string index $fline [incr i]]
# debug_out "ch1='$ch1' $ch1 != \"\"\" value=$value"
      if { $i >= $sl } break;
    }
    set fline [string range $fline $i end]
    set fline [string trimleft $fline]
    set sl [string length $fline]
    set key($vname) $value 
# debug_out "key($vname) == \"$value\""
  }
}

proc save_append { vname } {
  upvar $vname key
  if [info exists key] {
    return $key
  } else { 
    return "" 
  }
}

proc read_entry { stoken fline } {
# debug_out "read_entry $stoken $fline"
  set out_line {}
  set sidx [string length $stoken]
  set eidx [expr [string length $fline]-3]
  set fline [string range $fline $sidx $eidx]
  att_value_list dir_entry [string trim $fline]

  lappend out_line $dir_entry(name)
  lappend out_line [get_filetype $stoken $dir_entry(name)]
  lappend out_line [save_append dir_entry(size)]
  lappend out_line [format_date [save_append dir_entry(modified)]]
  lappend out_line [save_append dir_entry(user-perm)]
  lappend out_line [save_append dir_entry(group-perm)]

  return $out_line
}

proc obex_chkarg { cmd reason check } {
  debug_out "obex_chkarg $cmd $reason $check" 2
  if ![string compare $check ""] {
    internal_error "called obexftp '$cmd' without '$reason'"
  }
}

proc obexfileio { params } {
  set obexecute [getObexCfg obexftp command]
  
	set ret_val {}
  set pipecmd "|$obexecute $params"
debug_var pipecmd 3
  set infile [open $pipecmd "r"]
  fconfigure $infile -translation binary
### single file read is done
  set lines [read -nonewline $infile]
  catch {close $infile}
# debug_var lines 3 
### part, if no XML output (eg. --info command)
   if {[string first "<" $lines]<0} {
    foreach line [split $lines "\n"] {
      lappend ret_val [string trim $line]
    }
  } else {
    while {[set start [string first "<" $lines]]>=0} {
# debug_out "start=$start" 4
      set endst [string first ">" $lines]
      set nextl [string range $lines $start $endst]
      lappend ret_val [string trim $nextl]
      set lines [string range $lines [incr endst] end]
    }
  }

#  debug_out "obexfileio -> $ret_val" 4 
  return $ret_val
}

proc obexftp { cmd args } {
  global OBEXDIR
  variable obex0104_patch
debug_out "obexftp $cmd $args" 3

  set chdir [getObexCfg fileopt chdir]

  set fn1 [lindex $args 0]
  set fn2 [lindex $args 1]
  set fn3 [lindex $args 2]
  set ret_val {}

  case $cmd {
    ls {
      set cmd [getObexCfg fileopt list]
      obex_chkarg $cmd path $fn1
      set dir [file dirname $fn1]
      set nam [file tail $fn1]
####  set command "$cmd {$fn1}" 
      set command "$chdir {$dir} $cmd {$nam}"
    }
    gf { 
      set cmd [getObexCfg fileopt get]
      obex_chkarg $cmd pathname $fn1
      set dir [file dirname $fn1]
      set nam [file tail $fn1]
      set command "$chdir {$dir} $cmd {$nam}"
    }
    pf { 
      set cmd [getObexCfg fileopt put]
      obex_chkarg $cmd sourcename $fn1 
      obex_chkarg $cmd destination $fn2
      if $obex0104_patch {
        set dirname [file dirname $fn1]
        set filname [file tail $fn1]
        set currdir [pwd]
        cd $dirname 
        if {[string length $fn2]>1} {set fn2 [string trimleft $fn2 "/"]}
        set command "$chdir {$fn2} $cmd {$filname}" 
        cd $currdir
      } else {
        set command "$chdir {$fn2} $cmd {$fn1}" 
      }
    }
    mv { 
      set cmd [getObexCfg fileopt move]
      obex_chkarg $cmd sourcename $fn1
      obex_chkarg $cmd destination $fn2
      set command "$cmd {$fn1} {$fn2}" 
    }
    rm { 
      set cmd [getObexCfg fileopt delete]
      obex_chkarg $cmd pathname $fn1
      set dir [file dirname $fn1]
      set nam [file tail $fn1]
      set command "$chdir {$dir} $cmd {$nam}"
    }
    md { 
      set cmd [getObexCfg fileopt chdir]
      obex_chkarg $cmd dirname $fn1
      set command "$cmd {$fn1}" 
    }
    st { 
      ### This only work for Siemens :-(
      set cmd [getObexCfg fileopt info]
      obex_chkarg $cmd status "dummy"
      set command "$cmd" 
    }
    ve { 
      ### New option to retrieve obexftp version
      set cmd [getObexCfg fileopt version]
      obex_chkarg $cmd version "dummy"
      set command "$cmd" 
    }
    default {
      internal_error "Unknown command: '$cmd' $args"
    }
  }

### debug_var command 3
###   set infile [open $pipecmd "r"]
###   fconfigure $infile -translation binary
### ### single file read is done
###   set lines [read -nonewline $infile]
###   catch {close $infile}
### # debug_var lines 3 
### ### part, if no XML output (eg. --info command)
###    if {[string first "<" $lines]<0} {
###     foreach line [split $lines "\n"] {
###       lappend ret_val [string trim $line]
###     }
###   } else {
###     while {[set start [string first "<" $lines]]>=0} {
### # debug_out "start=$start" 4
###       set endst [string first ">" $lines]
###       set nextl [string range $lines $start $endst]
###       lappend ret_val [string trim $nextl]
###       set lines [string range $lines [incr endst] end]
###     }
###   }
### 
### #  debug_out "obexftp -> $ret_val" 4 
###   return $ret_val
  return [obexfileio $command]
}

proc list_dir { path } {
debug_out "list_dir $path" 4 
  variable separator 
  variable TAG 
  set ret_val ""
  set obex_dir [obexftp ls $path]

  set state begin
  foreach line $obex_dir {
# debug_out "line=$line" 4
    switch -- $state {
      begin {
        if {-1!=[string first $TAG(XML_identifier) $line]} {set state header}
      }
      header {
        if {-1!=[string first $TAG(DTD_identifier) $line]} {set state listing}
      }
      listing {
        if {-1!=[string first $TAG(folder-listbeg) $line]} {set state entities}
      }  
      end_listing {
        if {-1!=[string first $TAG(folder-listend) $line]} {set state end_data}
      }
      entities {
        set stoken [lindex [split $line $separator] 0]
        if {![string compare $stoken $TAG(folder_begin)]} {
          lappend ret_val [read_entry $stoken $line]
        } elseif {![string compare $stoken $TAG(file_begin)]} {
          lappend ret_val [read_entry $stoken $line] 
#### TAG(parent-folder) is used on Nokia 6230-phone ... do nozzin ...
#### patch from Michael Wikberg - 20040713
        } elseif {![string match $TAG(parent-folder)* $stoken]} {
        } else {
          set state end_listing
        }
      }
      end_data {
        debug_out "line after listing ignored ignored..." 3
      }
      default {
        debug_out "ignorin line: $line" 2
      }
    }
  }
  return $ret_val
}

proc read_file_tmp { path } {
  variable tmp_prefix
  debug_out "read_file_tmp $path" 3

  set renamed ""
  set name [file tail $path]
	set curr_dir [pwd]

  if [file exists $name] {
    set idx 0
    set fn $name.$idx
    while 1 { 
      if ![file exists $fn] break;
      incr idx
      set fn $name.$idx
      debug_out "$name.$idx -> [file exists $fn]" 6
    }
    set renamed $fn
    file rename $name $renamed
    debug_out "File '$name' temporarily renamed to '$renamed'" 
  }
  
  set ext [extension $path]
  while 1 {
    set local "$tmp_prefix[clock clicks].$ext"
    if ![file exists $local] break
    debug_out "temp file name collision '$local'"
  }
debug_var local 4 

  set cnt 3 
  
  obexftp gf $path 
  if ![file exists $name] {
    set msg "Unable to create file '%s' in current folder %s!"
    warning [format $msg $name $curr_dir]
    return {}
  }

debug_out "renaming '$name' to '$local'" 3
  file rename -force $name $local
  if {![string_empty $renamed]} {
    file rename $renamed $name
    debug_out "File renamed back to '$name'" 3
  }

  return $local
}

proc write_file_tmp { local path } {
  debug_out "write_file_tmp $local $path" 3
  set renamed ""

  set name [file tail $path]
  set dir_name [file dirname $path]
  set curr_dir [pwd]
	
  if [file exists $name] {
    set idx 0
    set fn $name.$idx
    while 1 { 
      if ![file exists $fn] break;
      incr idx
      set fn $name.$idx
      debug_out "$name.$idx -> [file exists $fn]" 6
    }
    set renamed $fn
    file rename $name $renamed
    debug_out "File '$name' temporarily renamed to '$renamed'" 
  }

  debug_out "File '$local' renamed to '$name'" 4
  file rename $local $name
  if ![file exists $name] {
    if ![string_empty $renamed] {
      file rename $renamed $name
      debug_out "File '$renamed' renamed back to '$name'" 3
    }
    set msg "Unable to create file '%s' in current folder %s!"
    warning [format $msg $path $curr_dir]
    return {}
  }

  obexftp pf $name $dir_name
  file delete $name 
  debug_out "Local file '$name' deleted" 3

  if {![string_empty $renamed]} {
    file rename $renamed $name
    debug_out "File renamed back to '$name'" 3
  }

}

proc path_exists { which path } {
debug_out "path_exists $which $path" 3

###
### check 1 complete path name
###
  set nam [file tail $path]
  set fnd [list_dir $path]
# debug_var fnd
### # if [llength $fnd] {return 1}
  foreach fil $fnd {
    set fnam [lindex $fil 0]
    set ftyp [lindex $fil 1]
debug_out "$fnam $ftyp" 3
    set isfile [string compare "folder" $ftyp]
    set nameeq ![string compare $nam $fnam]
debug_var isfile 3
debug_var nameeq 3
    switch $which {
      any  { if [expr $nameeq            ] {return 1} }
      dir  { if [expr $nameeq && !$isfile] {return 1} }
      file { if [expr $nameeq &&  $isfile] {return 1} }
      default {
        internal_error "path_exists: invalid path type '$which'!"
      }
    }
  }
###
### check 2 file/folder name in upper level folder
###
  set path [file dirname $path]
  set fnd [list_dir $path]
# debug_var fnd
#### if [llength $fnd] {return 1}
  foreach fil $fnd {
    set fnam [lindex $fil 0]
    set ftyp [lindex $fil 1]
    set isfile [string compare "folder" $ftyp]
    set nameeq ![string compare $nam $fnam]
debug_var isfile 3
debug_var nameeq 3
    switch $which {
      any  { if [expr $nameeq            ] {return 1} }
      dir  { if [expr $nameeq && !$isfile] {return 1} }
      file { return 0 }
      default {
        internal_error "path_exists: invalid path type '$which'!"
      }
    }
  }
  return 0
}
} ; ### end eval namespace
